@@ -1,12 +1,11 @@
 import base64
 import binascii
-import collections
-import crypt
 import errno
 import hashlib
 import logging
 import os
 import random
+import re
 import secrets
 import uuid
 import yaml
@@ -17,6 +16,11 @@ from numpy.random import bytes as np_random_bytes
 from python_secrets.utils import redact, find, prompt_string
 from python_secrets.google_oauth2 import GoogleSMTP
 from shutil import copy, copytree
+# >> Issue: [B404:blacklist] Consider possible security implications associated with run module.  # noqa
+#    Severity: Low   Confidence: High
+#    Location: python_secrets/secrets.py:21
+#    More Info: https://bandit.readthedocs.io/en/latest/blacklists/blacklist_imports.html#b404-import-subprocess  # noqa
+from subprocess import run, PIPE  # nosec
 from xkcdpass import xkcd_password as xp
 
 DEFAULT_SIZE = 18
@@ -33,6 +37,12 @@ SECRET_TYPES = [
         {'Type': 'uuid4', 'Description': 'UUID4 token'},
         {'Type': 'random_base64', 'Description': 'Random BASE64 token'}
         ]
+SECRET_ATTRIBUTES = [
+    'Variable',
+    'Type',
+    'Export',
+    'Prompt'
+]
 HOME = os.path.expanduser('~')
 CWD = os.getcwd()
 SECRETS_ROOT = os.path.join(
@@ -63,9 +73,12 @@ class SecretsEnvironment(object):
                  secrets_file="secrets.yml",
                  create_root=True,
                  defer_loading=True,
-                 source=None):
+                 export_env_vars=False,
+                 env_var_prefix=None,
+                 source=None,
+                 cwd=CWD):
         self._environment = environment \
-            if environment is not None else os.path.basename(CWD)
+            if environment is not None else os.path.basename(cwd)
         self._secrets_root = secrets_root
         self._secrets_file = secrets_file
         self._secrets_descriptions = "{}.d".format(
@@ -78,10 +91,28 @@ class SecretsEnvironment(object):
             else:
                 raise RuntimeError('Directory {} '.format(self.root_path()) +
                                    'does not exist and create_root=False')
+        self.export_env_vars = export_env_vars
+
+        # When exporting environment variables, include one that specifies the
+        # environment from which these variables were derived. This also works
+        # around a limitation in Ansible where the current working directory
+        # from which "ansible" was run. (The D2 lookup_plugin/python_secrets.py
+        # script needs to communicate this back to python_secrets in order for
+        # it's .python_secrets_environment file to be used to identify the
+        # proper environment.)
+
+        if self.export_env_vars is True:
+            os.environ['PYTHON_SECRETS_ENVIRONMENT'] = self._environment
+
+        self.env_var_prefix = env_var_prefix
+        # Secrets attribute maps; anything else throws exception
+        for a in SECRET_ATTRIBUTES:
+            setattr(self, a, dict())
         if source is not None:
             self.clone_from(source)
-        self._secrets = collections.OrderedDict()
-        self._descriptions = collections.OrderedDict()
+            self.read_secrets_descriptions()
+        self._secrets = dict()
+        self._descriptions = dict()
         self._changed = False
         self._groups = []
 
@@ -97,9 +128,39 @@ class SecretsEnvironment(object):
         """Create secrets root directory"""
         os.mkdir(self._secrets_root, mode=mode)
 
-    def environment_path(self):
-        """Returns the absolute path to secrets environment directory"""
-        return os.path.join(self._secrets_root, self._environment)
+    def environment_path(self, subdir=None, host=None):
+        """Returns the absolute path to secrets environment directory
+        or subdirectories within it"""
+        _path = os.path.join(self._secrets_root, self._environment)
+
+        if not (subdir is None and host is None):
+            valid_subdir = 'a-zA-Z0-9_/'
+            invalid_subdir = re.compile('[^{}]'.format(valid_subdir))
+            valid_host = 'a-zA-Z0-9_\./'
+            invalid_host = re.compile('[^{}]'.format(valid_host))
+
+            if subdir is None and host is not None:
+                raise RuntimeError('Must specify subdir when specifying host')
+
+            if subdir is not None:
+                if subdir.startswith('/'):
+                    raise RuntimeError('subdir may not start with "/"')
+                elif subdir.endswith('/'):
+                    raise RuntimeError('subdir may not end with "/"')
+                if not bool(invalid_subdir.search(subdir)):
+                    _path = os.path.join(_path, subdir)
+                else:
+                    raise RuntimeError('Invalid character in subdir: ' +
+                                       'must be in [{}]'.format(valid_subdir))
+
+            if host is not None:
+                if not bool(invalid_host.search(host)):
+                    _path = os.path.join(_path, host)
+                else:
+                    raise RuntimeError('Invalid character in host: ' +
+                                       'must be in [{}]'.format(valid_host))
+
+        return _path
 
     def environment_exists(self):
         """Return whether secrets environment directory exists
@@ -152,43 +213,80 @@ class SecretsEnvironment(object):
     def descriptions_path_create(self, mode=DEFAULT_MODE):
         """Create secrets descriptions directory"""
         if not self.environment_exists():
-            self.environment_path_create(mode=mode)
+            self.environment_create(mode=mode)
         if not self.descriptions_path_exists():
             os.mkdir(self.descriptions_path(), mode=mode)
 
+    def requires_environment(self):
+        """
+        Provide consistent error handling for any commands that require
+        an environment actually exist in order to work properly.
+        """
+        if not self.environment_exists():
+            raise RuntimeError(
+                'environment "{}" does not exist'.format(self._environment))
+
     def keys(self):
         """Return the keys to the secrets dictionary"""
         return [s for s in self._secrets.keys()]
 
     def items(self):
-        """Return the keys to the secrets dictionary"""
+        """Return the items from the secrets dictionary."""
         return self._secrets.items()
 
-    def get_secret(self, secret):
+    def get_secret(self, secret, allow_none=False):
         """Get the value of secret
 
         :param secret: :type: string
         :return: value of secret
         """
-        return self._secrets[secret]
+        if secret is None:
+            raise RuntimeError('Must specify secret to get')
+        v = self._secrets.get(secret, None)
+        if v is None and not allow_none:
+            raise RuntimeError('{} is not defined'.format(secret))
+        return v
 
-    def set_secret(self, secret, value):
-        """Set secret to value
+    def get_secret_export(self, secret):
+        """Get the specified environment variable for exporting secret
+
+        :param secret: :type: string
+        :return: environment variable for exporting secret
+        """
+        return self.Export.get(secret, None)
+
+    def _set_secret(self, secret, value):
+        """Set secret to value and export environment variable
 
         :param secret: :type: string
         :param value: :type: string
         :return:
         """
-        self._secrets[secret] = value
+        self._secrets[secret] = value  # DEPRECATED
+        getattr(self, 'Variable')[secret] = value
+        if self.export_env_vars:
+            _env_var = self.get_secret_export(secret)
+            if _env_var is None:
+                if self.env_var_prefix is not None:
+                    _env_var = '{}{}'.format(self.env_var_prefix, secret)
+                else:
+                    _env_var = secret
+            os.environ[_env_var] = str(value)
+
+    def set_secret(self, secret, value):
+        """Set secret to value and record change
+
+        :param secret: :type: string
+        :param value: :type: string
+        :return:
+        """
+        self._set_secret(secret, value)  # DEPRECATED
+        getattr(self, 'Variable', {secret: value})
         self._changed = True
 
     def get_type(self, variable):
         """Return type for variable or None if no description"""
-        for group in self._descriptions.keys():
-            for i in self._descriptions[group]:
-                if i['Variable'] == variable:
-                    return i['Type']
-        return None
+        return self.Type.get(variable, None)
 
     def changed(self):
         """Return boolean reflecting changed secrets."""
@@ -197,14 +295,51 @@ class SecretsEnvironment(object):
     def read_secrets_and_descriptions(self):
         """Read secrets descriptions and secrets."""
         self.read_secrets_descriptions()
-        self.read_secrets()
+        self.read_secrets(from_descriptions=True)
+        self.find_new_secrets()
 
-    def read_secrets(self):
-        """Load the current secrets from .yml file"""
+    def find_new_secrets(self):
+        """
+        Ensure that any new secrets defined in description files are
+        called out and/or become new undefined secrets.
+        :return:
+        """
+        # TODO(dittrich): Replace this with simpler use of attribute maps
+        for group in self._descriptions.keys():
+            for i in self._descriptions[group]:
+                s = i['Variable']
+                t = i['Type']
+                if self.get_secret(s, allow_none=True) is None:
+                    self.LOG.info('new {} '.format(t) +
+                                  'variable "{}" '.format(s) +
+                                  'is not defined')
+                    self._set_secret(s, None)
+
+    def read_secrets(self, from_descriptions=False):
+        """
+        Load the current secrets from .yml file
+
+        If no secrets have been set yet and from_descriptions is True,
+        return a dictionary comprised of the keys from the
+        descriptions dictionary defined to be None and set self._changed
+        to ensure these are written out.
+        """
         _fname = self.secrets_file_path()
         self.LOG.debug('reading secrets from {}'.format(_fname))
-        with open(_fname, 'r') as f:
-                self._secrets = yaml.safe_load(f)
+        try:
+            with open(_fname, 'r') as f:
+                _secrets = yaml.safe_load(f)
+            for k, v in _secrets.items():
+                self._set_secret(k, v)
+        except FileNotFoundError as err:
+            if from_descriptions:
+                for group in self._descriptions.keys():
+                    for i in self._descriptions[group]:
+                        self._set_secret(i['Variable'], None)
+                # Ensure these get written out to create a secrets file.
+                self._changed = True
+            else:
+                raise err
 
     def write_secrets(self):
         """Write out the current secrets for use by Ansible,
@@ -219,6 +354,7 @@ class SecretsEnvironment(object):
                           explicit_start=True,
                           default_flow_style=False
                           )
+            self._changed = False
         else:
             self.LOG.debug('not writing secrets (unchanged)')
 
@@ -227,6 +363,39 @@ class SecretsEnvironment(object):
         dest = self.environment_path()
         if source is not None:
             copyanything(source, dest)
+        self.read_secrets_descriptions()
+        self.find_new_secrets()
+
+    def get_descriptions(self, infile=None):
+        """
+        Read a secrets description file and return a dictionary if valid.
+
+        :param infile:
+        :return: dictionary of descriptions
+        """
+        with open(infile, 'r') as f:
+            data = yaml.safe_load(f)
+        for d in data:
+            for k in d.keys():
+                if k not in SECRET_ATTRIBUTES:
+                    raise RuntimeError('Invalid attribute ' +
+                                       '"{}"'.format(k) +
+                                       'in {}'.format(infile))
+        return data
+
+    def check_duplicates(self, data=list()):
+        """
+        Check to see if any 'Variable' dictionary elements in list match
+        any already defined variables. If so, raise RuntimeError().
+
+        :param data: list of dictionaries containing secret descriptions
+        :return: None
+        """
+        for d in data:
+            v = d.get('Variable')
+            if v in self._secrets:
+                raise RuntimeError('Variable "{}" '.format(v) +
+                                   'duplicates an existing variable')
 
     def read_secrets_descriptions(self):
         """Load the descriptions of groups of secrets from a .d directory"""
@@ -244,9 +413,26 @@ class SecretsEnvironment(object):
                 # dictionaries as dictionary keyed on group name.
                 for fname in file_names:
                     group = os.path.splitext(fname)[0]
-                    with open(os.path.join(groups_dir, fname), 'r') as f:
-                        data = yaml.safe_load(f)
-                    self._descriptions[group] = data
+                    data = self.get_descriptions(
+                        os.path.join(groups_dir, fname))
+                    if data is not None:
+                        self._descriptions[group] = data
+                        # Dynamically create maps keyed on variable name
+                        # for simpler lookups. (See the get_prompt() method
+                        # for an example.)
+                        for d in data:
+                            for k, v in d.items():
+                                try:
+                                    # Add to existing map
+                                    v = v if v != d['Variable'] else None
+                                    getattr(self, k)[d['Variable']] = v
+                                except AttributeError:
+                                    raise RuntimeError(
+                                        '"{}" is not '.format(k) +
+                                        'a valid attribute')
+                    else:
+                        raise RuntimeError('descriptions for group ' +
+                                           '"{}" is empty'.format(group))
             except Exception:
                 self.LOG.info('no secrets descriptions files found')
         else:
@@ -269,6 +455,10 @@ class SecretsEnvironment(object):
                     return None
         return None
 
+    def get_prompt(self, secret):
+        """Get the prompt for the secret"""
+        return self.Prompt.get(secret, secret)
+
     # TODO(dittrich): Not very DRY (but no time now)
     def get_secret_arguments(self, variable):
         """Get the Arguments of variable from set of secrets descriptions"""
@@ -373,6 +563,10 @@ def generate_password(unique=False):
 @Memoize
 def generate_crypt6(unique=False, password=None, salt=None):
     """Generate a crypt() style SHA512 ("$6$") digest"""
+    try:  # Python 3
+        import crypt
+    except ModuleNotFoundError as e:
+        raise
     if password is None:
         raise RuntimeError('generate_crypt6(): "password" is not defined')
     if salt is None:
@@ -470,6 +664,7 @@ class SecretsShow(Lister):
 
     def take_action(self, parsed_args):
         self.LOG.debug('showing secrets')
+        self.app.secrets.requires_environment()
         self.app.secrets.read_secrets_and_descriptions()
         variables = []
         if parsed_args.args_group:
@@ -480,13 +675,14 @@ class SecretsShow(Lister):
         else:
             variables = parsed_args.args \
                 if len(parsed_args.args) > 0 \
-                else [k for k in self.app.secrets._secrets.keys()]
-        columns = ('Variable', 'Type', 'Value')
+                else [k for k, v in self.app.secrets.items()]
+        columns = ('Variable', 'Type', 'Export', 'Value')
         data = (
                 [(k,
                   self.app.secrets.get_secret_type(k),
+                  self.app.secrets.get_secret_export(k),
                   redact(v, parsed_args.redact))
-                    for k, v in self.app.secrets._secrets.items()
+                    for k, v in self.app.secrets.items()
                     if k in variables]
         )
         return columns, data
@@ -573,7 +769,7 @@ class SecretsSet(Command):
             if '=' in arg:
                 k, v = arg.split('=')
             else:
-                k, v = arg, self.app.secrets.get_secret(arg)
+                k, v = arg, self.app.secrets.get_secret(arg, allow_none=True)
             k_type = self.app.secrets.get_type(k)
             if k_type is None:
                 self.LOG.info('no description for {}'.format(k))
@@ -581,7 +777,9 @@ class SecretsSet(Command):
                                    'has no description')
             if k_type == 'string':
                 if '=' not in arg:
-                    v = prompt_string(prompt=k, default=v)
+                    v = prompt_string(
+                        prompt=self.app.secrets.get_prompt(k),
+                        default=v)
                     if v is None:
                         self.LOG.info('no user input for "{}"'.format(k))
                         return None
@@ -592,10 +790,53 @@ class SecretsSet(Command):
                         _path = v[1:]
                     with open(_path, 'r') as f:
                         v = f.read().strip()
+                elif v.startswith('!'):
+                    # >> Issue: [B603:subprocess_without_shell_equals_true] subprocess call - check for execution of untrusted input.  # noqa
+                    #    Severity: Low   Confidence: High
+                    #    Location: python_secrets/secrets.py:641
+                    p = run(v[1:].split(),
+                            stdout=PIPE,
+                            stderr=PIPE,
+                            shell=False)
+                    v = p.stdout.decode('UTF-8').strip()
                 self.LOG.debug('setting {}'.format(k))
                 self.app.secrets.set_secret(k, v)
 
 
+class SecretsGet(Command):
+    """Get value associated with a secret"""
+
+    LOG = logging.getLogger(__name__)
+
+    def get_parser(self, prog_name):
+        parser = super().get_parser(prog_name)
+        parser.add_argument(
+            '-C', '--content',
+            action='store_true',
+            dest='content',
+            default=False,
+            help="Get content if secret is a file path " +
+            "(default: False)"
+        )
+        parser.add_argument('secret', nargs='?', default=None)
+        return parser
+
+    def take_action(self, parsed_args):
+        self.LOG.debug('get secret')
+        self.app.secrets.requires_environment()
+        self.app.secrets.read_secrets_and_descriptions()
+        if parsed_args.secret is not None:
+            value = self.app.secrets.get_secret(
+                parsed_args.secret, allow_none=True)
+            if not parsed_args.content:
+                print(value)
+            else:
+                if os.path.exists(value):
+                    with open(value, 'r') as f:
+                        content = f.read().replace('\n', '')
+                    print(content)
+
+
 class SecretsSend(Command):
     """
     Send secrets using GPG encrypted email.
@@ -658,33 +899,43 @@ class SecretsSend(Command):
         return parser
 
     def take_action(self, parsed_args):
+        self.LOG.debug('send secret(s)')
+        self.app.secrets.requires_environment()
+        self.app.secrets.read_secrets_and_descriptions()
         # Attempt to get refresh token first
         orig_refresh_token = None
-        try:
-            self.refresh_token = self.app.secrets[
-                "google_oauth_refresh_token"]
-        except KeyError:
-            pass
+        self.refresh_token =\
+            self.app.secrets.get_secret('google_oauth_refresh_token',
+                                        allow_none=True)
         if parsed_args.refresh_token:
             orig_refresh_token = self.refresh_token
             self.LOG.debug('refreshing Google Oauth2 token')
         else:
             self.LOG.debug('sending secrets')
+        if parsed_args.smtp_username is not None:
+            username = parsed_args.smtp_username
+        else:
+            username = self.app.secrets.get_secret(
+                'google_oauth_username')
         googlesmtp = GoogleSMTP(
-            parsed_args.smtp_username,
-            client_id=self.app.secrets['google_oauth_client_id'],
-            client_secret=self.app.secrets['google_oauth_client_secret'],
+            username=username,
+            client_id=self.app.secrets.get_secret(
+                'google_oauth_client_id'),
+            client_secret=self.app.secrets.get_secret(
+                'google_oauth_client_secret'),
             refresh_token=self.refresh_token
         )
         if parsed_args.refresh_token:
             new_refresh_token = googlesmtp.get_authorization()[0]
             if new_refresh_token != orig_refresh_token:
-                self.app.set_secret('google_oauth_refresh_token',
-                                    new_refresh_token)
+                self.app.secrets.set_secret('google_oauth_refresh_token',
+                                            new_refresh_token)
             return None
         elif parsed_args.test_smtp:
-            auth_string = googlesmtp.refresh_authorization()
-            googlesmtp.test_smtp(self, auth_string)
+            auth_string, expires_in = googlesmtp.refresh_authorization()
+            googlesmtp.test_smtp(
+                googlesmtp.generate_oauth2_string(
+                    base64_encode=True))
 
         recipients = list()
         variables = list()
@@ -692,23 +943,23 @@ class SecretsSend(Command):
             if "@" in arg:
                 recipients.append(arg)
             else:
-                if arg not in self.app.secrets:
-                    raise NameError(
-                        "Secret '{}' is not defined".format(arg))
-                variables.append(arg)
+                if self.app.secrets.get_secret(arg):
+                    variables.append(arg)
         message = "The following secret{} {} ".format(
             "" if len(variables) == 1 else "s",
             "is" if len(variables) == 1 else "are"
             ) + "being shared with you:\n\n" + \
             "\n".join(
-                ['{}={}'.format(v, self.app.secrets[v]) for v in variables]
+                ["{}='{}'".format(v, self.app.secrets.get_secret(v))
+                 for v in variables]
             )
-        # https://stackoverflow.com/questions/33170016/how-to-use-django-1-8-5-orm-without-creating-a-django-project/46050808#46050808
+        # https://stackoverflow.com/questions/33170016/how-to-use-django-1-8-5-orm-without-creating-a-django-project/46050808#46050808  # noqa
         for recipient in recipients:
             googlesmtp.send_mail(parsed_args.smtp_sender,
                                  recipient,
                                  parsed_args.smtp_subject,
                                  message)
+            self.LOG.info('sent encrypted secrets to {} '.format(recipient))
 
 
 class SecretsPath(Command):
@@ -718,7 +969,7 @@ class SecretsPath(Command):
 
     def get_parser(self, prog_name):
         parser = super(SecretsPath, self).get_parser(prog_name)
-        default_environment = self.app_args.environment
+        default_environment = self.app.options.environment
         parser.add_argument('environment',
                             nargs='?',
                             default=default_environment)
@@ -726,8 +977,9 @@ class SecretsPath(Command):
 
     def take_action(self, parsed_args):
         self.LOG.debug('returning secrets path')
-        e = parsed_args.environment
-        print(SecretsEnvironment(environment=e).secrets_file_path())
+        e = SecretsEnvironment(environment=parsed_args.environment)
+        e.requires_environment()
+        print(e.secrets_file_path())
 
 
 # vim: set fileencoding=utf-8 ts=4 sw=4 tw=0 et :
